import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from strapvizpy.bootstrap import calculate_boot_stats


def plot_ci(sample, rep, bin_size=30, n="auto", ci_level=0.95,
            ci_random_seed=None, title="", x_axis="Bootstrap Sample Mean", 
            y_axis="Count", path=None):
    
    """Makes a histogram of a boostrapped sampling distribution 
    with its confidence interval and oberserved mean.
     
    Parameters
    ----------
    sample : list or numpy.ndarray or pandas.core.series.Series 
        sample to bootstrap
    rep : int
        number of replicates of the distribution
    bin_size = int
        a number of bins representing intervals of equal size
        over the range
    n : str or int, default="auto"
        bootstrap sample size, "auto" specifies using the
        same size as the sample
    ci_level : float, default=0.95
        confidence level
    ci_random_seed : None or int, default=None
        seed for random state
    title : str, default = ""
        title of the histogram
    x_axis : str, default = "Bootstrap Sample Mean"
        name of the x axis
    y_axis : str, default = "Count"
        name of the y axis
    path : None or str, default = None
        specify the directory to save the figure as .png
    
    Returns
    -------
    plot: histogram
        histogram of bootstrap distribution with confidence interval
        and oberserved mean
    
    Examples
    --------
    >>> plot_ci([1, 2, 3, 4, 5, 6, 7], 1000, n=100, ci_level=0.95,
                ci_random_seed=123, title="Bootstrap")
    """

    if not isinstance(title, str):
        raise TypeError(
            "The value of the argument 'title' must be type of str."
        )
        
    if not isinstance(x_axis, str):
        raise TypeError(
            "The value of the argument 'x_axis' must be type of str."
        )
        
    if not isinstance(y_axis, str):
        raise TypeError(
            "The value of the argument 'y_axis' must be type of str."
        )

    if not (isinstance(path, str) or path == None):
        raise TypeError(
            "The value of the argument 'path' must be type of str or None."
        )
        
    if path is not None :
        if os.path.isdir(path) is False:
            raise NameError("The folder path you specified is invalid.")

    sample_stat_dict = calculate_boot_stats(sample, rep, level=ci_level, 
                                            n=n, random_seed = ci_random_seed,
                                            pass_dist=True)
        
    plt.hist(sample_stat_dict[1], density=False, bins=bin_size)
    plt.axvline(sample_stat_dict[0]["lower"], color='k', linestyle='--')
    plt.axvline(sample_stat_dict[0]["sample_mean"], color='r', linestyle='-')
    plt.axvline(sample_stat_dict[0]["upper"], color='k', linestyle='--')
    axes = plt.gca()
    _, y_max = axes.get_ylim()
    plt.text(sample_stat_dict[0]["sample_mean"], 
             y_max * 0.9 , 
             (str(round(sample_stat_dict[0]["sample_mean"], 2))+
              '('+u"\u00B1"+str(round(sample_stat_dict[0]['std_err'],2))+')'), 
             ha='center', va='center',rotation='horizontal', 
             color = "k", bbox={'facecolor':'white', 'pad':5})
    plt.text(sample_stat_dict[0]["upper"], 
             y_max * 0.9 , 
             (str(round(sample_stat_dict[0]["upper"], 2))), 
             ha='center', va='center',rotation='horizontal', 
             color = "k", bbox={'facecolor':'white', 'pad':5})
    plt.text(sample_stat_dict[0]["lower"], 
             y_max * 0.9 , 
             (str(round(sample_stat_dict[0]["lower"], 2))), 
             ha='center', va='center',rotation='horizontal', 
             color = "k", bbox={'facecolor':'white', 'pad':5})
    plt.title(title)
    plt.xlabel(x_axis)
    plt.ylabel(y_axis)

    if path is not None:
        plt.savefig(f"{path}bootstrap_histogram.png")

    return plt
    


def tabulate_stats(stat, precision=2, estimator=True, alpha=True, path=None):
    """Makes two tables that summerize the statistics from the bootstrapped 
    samples and the parameters for creating the bootstrapped samples. It also allows you
    to save the tables in html format. 


    Parameters
    ----------
    stat : dict or tuple
        summary statistics produced by the `calculate_boot_stats()` function 
    precision : int, default=2
        the precision of the table values
    estimator : boolean, default=True
        include the bootstrap estimate in the summary statistics table
    alpha : boolean, default=True
        include the significance level in the summary statistics table
    path : str, default = None
        specify a path to where the tex files of tables should be saved.

    Returns
    -------
    tuple :
        summary statistics: style object
            table summerizing the lower bound and upper bound of the confidence
            interval,the standard error, the sampling statitic (if estimator = True),
            and the significance level (if alpha = True). Style objects do not display
            well in a python shell.
        bootstrap parameters: style object
            table  summerizing the parameters of the bootstrap sampling spficiying
            the original sample size, number of repititions, the significance level,
            and the number of samples in each bootstrap if its different from the
            original sample size. Style objects do not display well in a python shell.
        
    Examples
    --------
    >>> st = calculate_boot_stats([1, 2, 3, 4], 1000, level=0.95, random_seed=123)
    >>> stats_table, parameter_table  = tabulate_stats(st)
    >>> stats_table
    >>> parameter_table
    """

    if not(isinstance(stat, tuple) | isinstance(stat, dict)):
        raise TypeError(
            "The stats parameter must be created from "
            "calculate_boot_stats() function."
        )
    if not isinstance(precision, int):
        raise TypeError("The precision parameter must be of type int.")
    if not (isinstance(estimator, bool) & isinstance(alpha, bool)):
        raise TypeError(
            "The estimator and alpha parameters must be of type boolean."
        )
    if not (isinstance(path, str) or path is None):
        raise TypeError("The path parameter must be a character string.")
        
    if path is not None :
         if os.path.isdir(path) is False:
            raise NameError("The folder path you specified is invalid.")
    
    if isinstance(stat, tuple):
        stat = stat[0]
        
    dic_keys = stat.keys()
    
    if not (("lower" in dic_keys) &
            ("upper" in dic_keys) &
            ("std_err" in dic_keys) &
            ("estimator" in dic_keys) &
            ("level" in dic_keys) &
            ("sample_size" in dic_keys) &
            ("n" in dic_keys) &
            ("rep" in dic_keys)):
        raise TypeError(
            "The statistics dictionary is missing a key. "
            "Please rerun calculate_boot_stats() function"
        )
        
    # define the statistics table
    df = pd.DataFrame(data=np.array([(stat["lower"], stat["upper"],
                                      stat["std_err"])]),
                      columns=["Lower Bound CI", "Upper Bound CI",
                               "Standard Error"])

    if estimator is True:
        s_name = "Sample " + stat["estimator"]
        df[s_name] = stat["sample_" + stat["estimator"]]

    if alpha is True:
        df["Significance Level"] = 1 - stat["level"]
        stats_table = df.style.format(
            precision=precision, formatter={("Significance Level"): "{:.3f}"}
        )
    else:
        stats_table = df.style.format(precision=precision)

    stats_table = stats_table.hide(axis="index")

    # set formatting and caption for table
    stats_table.set_caption(
        "Bootstrapping sample statistics from sample with "+
        str(stat["sample_size"]) + " records"
    ).set_table_styles(
        [{"selector": "caption",
          "props": "caption-side: bottom; font-size: 1.00em;"}],
        overwrite=False)
            
    # create bootstrapping parameter summary table
    df_bs = pd.DataFrame(
        data=np.array(
            [(stat["sample_size"], stat["rep"], (1 - stat["level"]))]),
        columns=["Sample Size", "Repetition", "Significance Level"])
    
    if stat["n"] != "auto":
        df_bs["Samples per bootstrap"] = round(stat["n"], 0)

    # set formatting and caption for table
    bs_params = df_bs.style.format(
        precision=0,
        formatter={("Significance Level"): "{:.3f}"}
    ).hide(axis="index")
    
    (bs_params
    .set_caption("Parameters used for bootstrapping")
    .set_table_styles(
        [{"selector": "caption",
          "props": "caption-side: bottom;font-size:1.00em;"}],
        overwrite=False)
    )

    if path is not None:
        with open(f"{path}sampling_statistics.tex", "w") as tf:
            tf.write(stats_table.to_latex())

        with open(f"{path}bootstrap_params.tex", "w") as tf:
            tf.write(bs_params.to_latex())
        
    return stats_table, bs_params



import numpy as np
import pandas as pd
import warnings

# Global constant for supported estimators
SUPPORTED_ESTIMATORS = {
    "mean": np.mean,
    "median": np.median,
    "var": np.var,
    "sd": np.std
}

def bootstrap_distribution(sample, rep, n="auto", estimator="mean", random_seed=None):
    """Bootstraps a sampling distribution for a sample.

    A sampling distribution of `rep` replicates is generated
    for the specified `estimator`with replacement with a
    bootstrap sample size of `n`.

    Parameters
    ----------
    sample : list or numpy.ndarray or pandas.core.series.Series 
        sample to bootstrap
    rep : int
        number of replicates of the distribution
    n : str or int, default="auto"
        bootstrap sample size, "auto" specifies using the same size as the sample
    estimator : {"mean", "median", "var", "sd"}
        sampling distributor's estimator
    random_seed : None or int, default=None
        seed for random state
    
    Returns
    -------
    numpy.ndarray
        bootstrapped sampling distribution
    
    Examples
    --------
    >>> bootstrap_distribution([1, 2, 3], 3, 3)
    array([1.66, 2, 2.66])
    """

    if not (isinstance(sample, list) or
            isinstance(sample, np.ndarray) or
            isinstance(sample, pd.Series)):
        raise TypeError("sample should be one of the types"
                        "[list, numpy.ndarray, pandas.core.series.Series]")

    if not isinstance(rep, int):
        raise TypeError("rep should be of type 'int'")

    if isinstance(rep, int) and rep < 1:
        raise ValueError("Invalid value for rep")

    if not (isinstance(n, str) or isinstance(n, int)):
        raise TypeError("n should be of type 'str' or 'int'")

    if isinstance(n, str) and n != "auto":
        raise ValueError("Invalid value for n. Did you intend n='auto'?")

    if isinstance(n, int) and n < 1:
        raise ValueError("Invalid value for n")

    if not isinstance(estimator, str):
        raise TypeError("estimator should be of type 'str'")

    if estimator not in SUPPORTED_ESTIMATORS.keys():
        raise ValueError("Supported estimators are mean, median, var, sd")

    if not (random_seed is None or isinstance(random_seed, int)):
        raise TypeError("random_seed should be None or of type 'int'")

    if isinstance(random_seed, int) and random_seed < 0:
        raise ValueError("Invalid value for random_seed")

    if random_seed:
        np.random.seed(random_seed)

    if n == "auto":
        n = len(sample)

    return SUPPORTED_ESTIMATORS[estimator](
        np.random.choice(sample, size=(rep, n), replace=True),
        axis=1
    )

def calculate_boot_stats(sample, rep, n="auto", level=0.95, estimator="mean", random_seed=None, pass_dist=False):
    """Calculates a bootstrapped confidence interval for a sample.

    A bootstrapped confidence interval for the desired estimator for
    the provided sample is calculated for a confidence level `level`.
    Other stats and parameters of the distribution and sample are
    also returned. 

    Parameters
    ----------
    sample : list or numpy.ndarray or pandas.core.series.Series 
        sample to bootstrap
    rep : int
        number of replicates of the distribution
    n : str or int, default="auto"
        bootstrap sample size, "auto" specifies using the same size as the sample
    level : float, default=0.95
        confidence level
    estimator : {"mean", "median", "var", "sd"}
        sampling distributor's estimator
    random_seed : None or int, default=None
        seed for random state
    pass_dist : bool, default = "False"
        return the bootstrapped sample distribution - False or True 
    
    Returns
    -------
    dictionary
        Dictionary containing lower and upper bootstrapped confidence
        interval for the desired estimator, along with the given estimator.
        Also 
    
    Examples
    --------
    >>> calculate_boot_stats([1, 2, 3, 4], 1000, level=0.95, random_seed=123)  
    {'lower': 1.5,
    'upper': 3.5,
    'sample_mean': 2.5,
    'std_err': 0.5414773771820943,
    'level': 0.95,
    'sample size': 4,
    'n': 'auto',
    'rep': 1000,
    'estimator': 'mean'}
    """

    if not isinstance(level, float):
        raise TypeError("level should be of type 'float'")

    if not (level > 0 and level < 1):
        raise ValueError("level should be between 0 and 1")

    if not isinstance(pass_dist, bool):
        raise TypeError("pass_dist should be of type 'bool'")

    if level < 0.7:
        warnings.warn("Warning: chosen level is quite \
            low--level is a confidence level, not a signficance level")

    # get the bootstrapped mean vector
    dist = bootstrap_distribution(sample=sample,
                                  rep=rep,
                                  n=n,
                                  estimator=estimator,
                                  random_seed=random_seed)

    stats_dict = {}

    stats_dict["lower"] = np.percentile(dist, 100 * (1-level)/2)
    stats_dict["upper"] = np.percentile(dist, 100 * (1-(1-level)/2))
    stats_dict["sample_" + estimator] = SUPPORTED_ESTIMATORS[estimator](sample)
    stats_dict["std_err"] = np.std(dist)
    stats_dict["level"] = level
    stats_dict["sample_size"] = len(sample)
    stats_dict["n"] = n
    stats_dict["rep"] = rep
    stats_dict["estimator"] = estimator

    if pass_dist:
        return stats_dict, dist
    else:
        return stats_dict



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['strapvizpy']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.7.1,<5.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.1,<2.0.0',
 'pandas>=1.4.0,<2.0.0']

setup_kwargs = {
    'name': 'strapvizpy',
    'version': '0.2.1',
    'description': 'Performs bootstrapping of a dataset to produce plots and statistics for use in final reports and documents. ',
    'long_description': "# StrapvizPy\n\n![example workflow](https://github.com/UBC-MDS/strapvizpy/actions/workflows/ci-cd.yml/badge.svg)\n[![codecov](https://codecov.io/gh/UBC-MDS/strapvizpy/branch/main/graph/badge.svg?token=ufgX4eYuYU)](https://codecov.io/gh/UBC-MDS/strapvizpy)\n[![Documentation Status](https://readthedocs.org/projects/strapvizpy/badge/?version=latest)](https://strapvizpy.readthedocs.io/en/latest/?badge=latest)\n\n## Summary\n\nPerforms bootstrapping of a dataset column to produce plots and statistics for use in final reports and documents.\n\nThe purpose of this package is to simplify and automate the process of creating simple bootstrap distributions of numerical data columns. The package will have a module which intakes a dataset column and relevant parameters such as the desired confidence bounds and number of simulations. The module will perform the simulation statistics to generate the bootstrap mean distribution and relevant statistics such as the sample mean and bootstrapped confidence interval. The package will also contain a module for visualization of the bootstraped confidence interval, and a module for creating a professional publication-ready table of the relevant statistics.\n\n## Package context within the Python ecosystem\n\nThe package will likely build on scipy's [stats module](https://docs.scipy.org/doc/scipy/reference/stats.html), which allows one to conduct the boostrap sampling in the first place using the bootstrap method. strapvizpy will streamline and extend this process from the pure statistical process done in this module. sklearn has a utils module with a [resample](https://scikit-learn.org/stable/modules/generated/sklearn.utils.resample.html) method which also seems popular and achieves similar functionality. While we cannot be certain that one does not exist, there does not seem to be a package which streamlines the process from data to visualization and presentation as proposed in this document. Some tutorials on bootstrap confidence intervals from [machinelearningmastery.com](https://machinelearningmastery.com/calculate-bootstrap-confidence-intervals-machine-learning-results-python/) and [towardsdatascience.com](https://towardsdatascience.com/bootstrapping-using-python-and-r-b112bb4a969e) encourage the reader to plot the results manually.\n\n\n## Installation\n\n```bash\n$ pip install strapvizpy\n```\n\n## Usage\n\nTo import strapvizpy and check the version:\n\n```python\nimport strapvizpy\nprint(strapvizpy.__version__)\n```\n\nTo import the suite of functions:\n\n```python\nfrom strapvizpy import bootstrap\nfrom strapvizpy import display\n```\n\nPlease view our packaged documentation [here](https://strapvizpy.readthedocs.io/en/latest/).\n\n## Functions\n\n- `bootstrap_distribution`: Returns a sampling distribution of specified replicates is generated for a specified estimator with replacement for a given bootstrap sample size.  \n- `calculate_boot_stats`: Calculates a confidence interval for a given sampling distribution as well as other bootstrapped statistics.  \n- `plot_ci`: Creates a histogram of a bootstrapped sampling distribution with its confidence interval and observed sample statistic.  \n- `tabulate_stats`: Generates a table that contains a given sampling distribution's mean and standard deviation along with relevant statistics as well as a summary table of the bootstrap distributions parameters. The code automatically saves the tables as html documents.\n\n## Contributing\nJulien Gordon, Gautham Pughazhendhi, Zack Tang, and Margot Vore.\n\n## License\n\n`StrapvizPy` was created by Julien Gordon, Gautham Pughazhendhi, Zack Tang, Margot Vore. It is licensed under the terms of the MIT license.\n\n## Credits\n\n`StrapvizPy` was created with [`cookiecutter`](https://cookiecutter.readthedocs.io/en/latest/) and the `py-pkgs-cookiecutter` [template](https://github.com/py-pkgs/py-pkgs-cookiecutter).\n",
    'author': 'Julien Gordon, Gautham Pughazhendhi, Zack Tang, Margot Vore',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
